'Unix time driver for the DS1307 RTC
'Copyright 2007 Raymond Allen
'Thanks to Mike Green and James Burrows for their I2C and DS1307 codes posted in "Object Exchange"
'NOTE:  Time zones implementation is fairly weak... Only has data for continental US in normal zones...
'Note:  Also suffers from Year 2038 problem, just like 32-bit UNIX...


CON  'Constants Section

   DS1307 = $D0   'Address of DS1307 for read mode (this plus 1 is address for write mode)
   
   ACK      = 0                        ' I2C Acknowledge
   NAK      = 1                        ' I2C No Acknowledge
   Xmit     = 0                        ' I2C Direction Transmit
   Recv     = 1                        ' I2C Direction Receive
   SCL      = 28             'I2C clock pin#
   SDA      = 29             'I2C data pin#

   SecondsPerHour = 60*60
   SecondsPerDay = SecondsPerHour*24
   

PUB StrTime(nFormat):strptr|p,s,pTime,i
  'Returns time as a formated time string
  'make the string
  s:=@strBuffer
  pTime:=@Time2
  case nFormat
    1:  
      'Local Time Format:  HH:MM:SS MM/DD/YYYY
      LocalTime(time)       'get local time
      s:=Dec(word[pTime][2],2,s)
      cout(":",s++)      
      s:=Dec(word[pTime][1],2,s)
      cout(":",s++)
      s:=Dec(word[pTime],2,s)
      cout(" ",s++)
      s:=Dec(word[pTime][4]+1,2,s)
      cout("/",s++)
      s:=Dec(word[pTime][3],2,s)
      cout("/",s++)
      s:=Dec(word[pTime][5]+1900,4,s)
      cout(" ",s++)
      s:=str(@tzname,s)
      cout(0,s++) 'terminate string
    2:
      'UNIX TIME (GMT)
      s:=Dec(MkTime(pTime),10,s)
      cout(0,s++) 'terminate string
    3:
      'Local time in C format
      LocalTime(time)       'get local time
      i:=@DaysOfWeek+word[pTime][6]*3
      repeat 3
        cout(byte[i++],s++)
      cout(" ",s++)
      i:=@Months+word[pTime][4]*3
      repeat 3
        cout(byte[i++],s++)
      cout(" ",s++)
      s:=Dec(word[pTime][3],2,s)
      cout(" ",s++)     
      s:=Dec(word[pTime][2],2,s)
      cout(":",s++)      
      s:=Dec(word[pTime][1],2,s)
      cout(":",s++)
      s:=Dec(word[pTime],2,s)
      cout(" ",s++)
      s:=Dec(word[pTime][5]+1900,4,s)
      cout(0,s++) 'terminate string   '

  return @strBuffer    
    
PUB SetTimeZone(n)  'n is hours west of GMT (outside of DST)
  case n
    5:
      bytemove(@tzname0,String("EST"),3)
      bytemove(@tzname1,String("EDT"),3) 
      timezone:= SecondsPerHour*5      'EST
      daylight:= SecondsPerHour        'EDT offset
    6:
      bytemove(@tzname0,String("CST"),3)
      bytemove(@tzname1,String("CDT"),3) 
      timezone:= SecondsPerHour*6      'EST
      daylight:= SecondsPerHour        'EDT offset
    7:
      bytemove(@tzname0,String("MST"),3)
      bytemove(@tzname1,String("MDT"),3) 
      timezone:= SecondsPerHour*7      'EST
      daylight:= SecondsPerHour        'EDT offset
    8:
      bytemove(@tzname0,String("PST"),3)
      bytemove(@tzname1,String("PDT"),3) 
      timezone:= SecondsPerHour*8      'EST
      daylight:= SecondsPerHour        'EDT offset

    
PRI Dec(i,p,s):splus|n,d,j 'print decimal number i with p digits to string buffer @ s
  if (i<0)
    cout("-",s++)
    i:=-i
  d:=1
  repeat (p-1)
    d*=10
  repeat p
    if (i=>d)
      n:=i/d
      i-=n*d
    else
      n:=0
    cout("0"+n,s++)
    d/=10
  return s  

PRI cOut(c,s)
   'prints a char to string buffer s
   byte[s]:=c

PRI Str(stringptr,s):splus
  'prints a string to string buffer s
  repeat strsize(stringptr)
    cout(byte[stringptr++],s++)
  return s   

PUB SetUnixTime(timeT)
   GmTime(timeT,@time1)
   SetTime1307


PUB SetGmtTime(month, day, year, hours, minutes,seconds):bOk
  'Set DS1307 to GMT time using user value ranges:
  'hours [0..23] 24-hour clock
  'minutes [0..59]
  'seconds [0..59]
  'month [1..12]
  'day [1..31]
  'year [2000-2099]
  'Does not check day to make sure it's valid (e.g., Feb 31 does not fail)
  If (month<1) OR (month>12) or (day<1) or (day>31) or (year<2000) or (year>2099)
    return false
  tm_mday:=day
  tm_mon:=month-1
  tm_year:=year-1900
  tm_wday:=DayOfWeek(month,day,year)
  tm_hour:=hours
  tm_min:=minutes
  tm_sec:=seconds
  SetTime1307   


PUB SetLocalTime(month, day, year, hours, minutes, seconds):bOk|posix
  'note that certian local times with an hour of DST boundaries can't be set or are ambiguous..
  If (month<1) OR (month>12) or (day<1) or (day>31) or (year<2000) or (year>2099)
    return false
  tm_mday2:=day
  tm_mon2:=month-1
  tm_year2:=year-1900
  tm_wday2:=DayOfWeek(month,day,year)
  tm_yday2:=DayOfYear(tm_year2,tm_mon2,tm_mday2)
  tm_hour2:=hours
  tm_min2:=minutes
  tm_sec2:=seconds
  posix:=MkTime(@Time2)'-GetTz(@Time2)+MkTime(@Time2)
  GmTime(posix,@Time1)'LocalTime(posix,@Time1)
  SetTime1307

     
PRI MkTime(pTime):secs 
  'converts local time in time structure to GMT seconds since Midnight Jan 1, 1970
  'Note that due to DST, one hour in local time spring doesn't exist and one in fall exists twice
  'inverse of LocalTime
  'Get seconds assuming pTime is GMT 
  secs:=Time2Secs(pTime)
  'adjust for time zone so result is in GMT
  secs+=GetTz(pTime)
  return secs
  

PUB Time:secs
  'return unix time from 1307
  GetTm1307
  return Time2Secs(@Time1)

PRI Time2Secs(pTime):secs|year,month,YearDay,i
  'converts time structure to epoch (assuming GMT)
  year:=word[pTime][5]+1900 'get year from time structure
  secs:=0
  repeat i from 1970 to (year-1)
    if (IsLeap(i))
      secs+=SecondsPerDay*366
    else
      secs+=SecondsPerDay*365 
  
  secs+=word[pTime][7]*SecondsPerDay  'add in day of year
  secs+=word[pTime][2]*SecondsPerHour  'add in hour
  secs+=word[pTime][1]*60   'add in minutes
  secs+=word[pTime]
  return secs
  
    
PRI LocalTime(timeT):bOk|ttime,pTime
  'convert #seconds since Midnight Jan 1, 1970 GMT to local time structure @Time2
  pTime:=@Time2
  'inverse of MkTime
  '1st, guess that not in DST and adjust for time zone
  ttime:=timeT-timezone'+daylight
  GmTime(ttime,pTime)
  if (MkTime(pTime)<>timeT)
    'must not be in daylight savings time
    ttime+=daylight
    GmTime(ttime,pTime)
    if (MkTime(pTime)<>timeT)
       'must be the double hour in November...  MkTime will presume Standard time after 2, but may have been given first hour of double hour
       'In this case, MkTime will give daylight more
       if (MkTime(pTime)==(timeT+daylight))
         'Fix the TZ info
         bytemove(@tzname,@tzname1,3)
       else
          return -1
  return 0           

PRI GmTime(secs,pTime)|year,i,yday, hour, minute,p, month,mday
  'convert unix time (#seconds since epoch (jan 1, 1970)) to GMT time structre @pTime
  
  'first, get the year
  year:=1970
  repeat 
    if (IsLeap(year))
      i:=366
    else
      i:=365      
    if (secs=>SecondsPerDay*i)
      secs-=SecondsPerDay*i
      year++
    else
      quit  'this is the year
  'next get the day of year
  yday:=secs/SecondsPerDay
  secs//=SecondsPerDay
  'next, get hour
  hour:=secs/SecondsPerHour
  secs//=SecondsPerHour
  'next, get minutes and seconds
  minute:=secs/60
  secs//=60

  'Now, write to time structure
  word[pTime]:=secs
  word[pTime][1]:=minute
  word[pTime][2]:=hour
  'calc month
  if (IsLeap(year))
    p:=@LeapDaysBeforeMon
  else
    p:=@NormDaysBeforeMon
  month:=0
  repeat  
    if (yday>word[p][month+1])
      month++
    else
      mday:=yday-word[p][month]+1
      quit
  word[pTime][3]:=mday  
  word[pTime][4]:=month
  word[pTime][5]:=year-1900
  word[pTime][6]:=DayOfWeek(month+1,mday,year)
  word[pTime][7]:=yday
  word[pTime][8]:=-1  'don't know DST status
    


PRI GetTZ(pTime):tz|i,standard,year,mday,hour,mon,bSaving
  'returns local time offset from GMT (in seconds) (based on local time in time structure @Time1)
  'Sets the tzname string based on whether or not DST is in affect
  'In USA with 1 hour DST:  Note that one local hour in Spring doesn't exist and one hour in Fall comes twice... This function presumes the Fall hour is not DST! 
  'Here's some relevant rules for North America (from txdata2001i.tar.gz)  'You Europeans are on your own! :)
    '# Rule  NAME    FROM    TO      TYPE    IN      ON      AT      SAVE    LETTER/S
    'Rule    US      1967    1973    -       Apr     lastSun 2:00    1:00    D
    'Rule    US      1974    only    -       Jan     6       2:00    1:00    D
    'Rule    US      1975    only    -       Feb     23      2:00    1:00    D
    'Rule    US      1976    1986    -       Apr     lastSun 2:00    1:00    D
    'Rule    US      1987    2006    -       Apr     Sun>=1  2:00    1:00    D
    'Rule    US      2007    max     -       Mar     Sun>=8  2:00    1:00    D
    'Rule    US      2007    max     -       Nov     Sun>=1  2:00    0       S
     
    '# Zone  NAME            GMTOFF  RULES   FORMAT  [UNTIL]
    'Zone    EST5EDT          -5:00  US      E%sT
    'Zone    CST6CDT          -6:00  US      C%sT
    'Zone    MST7MDT          -7:00  US      M%sT
    'Zone    PST8PDT          -8:00  US      P%sT

    bSaving:=false
    'using formula from here:  http://www.skipgeel.com/dst.html
    year:=word[pTime][5]+1900
    mday:= word[pTime][3]
    hour:= word[pTime][2]
    mon:=word[pTime][4]    
    'Appears to be valid from 1987
    if (year=>2007) 
      case (mon)
        2:  'DST begins in March on second Sunday at 2 AM
          i:=14-(1+year*5/4)//7
          if (mday>i) OR ((mday==i) AND (hour>1))
            bSaving:=true
        3..9:  'April to October is DST
          bSaving:=true
        10:  'DST ends in November on first sunday at 2 AM
          i:=7-(1+year*5/4)//7
          if NOT ((mday>i) OR ((mday==i) AND (hour>1)))
            bSaving:=true
    else
      case (mon)
        3:  'DST begins in April
          i:=1+(2+6*year-year/4)//7
          if (mday>i) OR ((mday==i) AND (hour>1))
            bSaving:=true
        3..8:  'April to September is DST
          bSaving:=true
        9:  'DST ends in October
          i:=31-(year*5/4)//7
          if NOT((mday>i) OR ((mday==i) AND (hour>1)))
            bSaving:=true

    
    if (bSaving)
      bytemove(@tzname,@tzname1,3)  
      return timezone-daylight
    else
      bytemove(@tzname,@tzname0,3)
      return timezone

PUB IsLeap(year):bLeap
  'Determines if year (ordinary integer, e.g., 2007, is a leap year or not
  'a year is leap if it is divisible by 4, unless it is divisible by 100, except when divisible by 400
  'So, 1900 is not a leap year, but 2000 is...
  return ((year// 4 == 0) AND ((year//100) <> 0)) OR ((year//400) == 0)  

  

PRI SetTime1307
  'Set the DS1307 Clock to GMT
  'Input is time structure @Time1
  'Uses DS1307 time structure @DS1307time
  DS1307day:=tm_wday 
  DS1307date:= ToBCD(tm_mday)  '[1..31] 
  DS1307month:=ToBCD(tm_mon+1)    'now [1..12]
  DS1307year:=ToBCD(tm_year-100)  'now [00..99] for years 2000 to 2099
  DS1307seconds:=ToBCD(tm_sec)& $7F  'making sure not to halt the clock! 
  DS1307minutes:= ToBCD(tm_min)  
  DS1307hours:=ToBCD(tm_hour) 
  OUT($00, 7, @DS1307seconds) 'write to DS1307 over i2c

'DS1307seconds byte 0          'Address 0H                  ' CH   |      10 Seconds    |          Seconds          |   00-59
'DS1307minutes byte 0          'Address 1H                  '      |      10 Minutes    |          Minutes          |   00-59 
'DS1307hours   byte 0          'Address 2H                  '12/24 |  0   |10/PM | Hour |            HOUR           |   01-12 or 00-23 

PUB ToBCD(i)  'convert binary byte to BCD
  return ((i/10)<<4) + (i//10)
  

PUB DayOfWeek(month, day, year):dow  '[0..6]
  'calculate day of the week using Zeller's formula
  'month [1..12], day [1..31]
  'Sunday==0
  if (month<3)
    year--
  dow:=year+year/4-year/100+year/400+byte[@Zeller+month-1]+day
  dow//=7
  'dow++
  return dow
    
PRI GetTm1307:bOk|i,j,k
  'read entire time structure from DS1307 (assumed to be GMT) and store at address DS1307time
  'then convert to time structure (Still in GMT) and store at address Time1
  bOk:=in(0, 8, @DS1307tm) 'gets whole time structure from DS1307
  if (NOT bOk)
    return false
  'seconds
  i:=byte[@DS1307tm+0]
  tm_sec:=(i>>4 & $07)*10 + (i & $0F)
  'minutes
  i:=byte[@DS1307tm+1]
  tm_min:=(i>>4 & $07)*10 + (i & $0F)
  'hours
  i:=byte[@DS1307tm+2]
  case (i & $40)
    0: '24 Hour mode
      tm_hour:=(i>>4 & $03)*10 + (i & $0F)
    other: '12 hour mode
      tm_hour:= (i>>4 & $01)*10 + (i & $0F) + 12*((i&$20)>>5)
  'day of week
  i:=byte[@DS1307tm+3] - 1
  tm_wday:=i
  'day of month
  i:=byte[@DS1307tm+4]
  tm_mday:=(i>>4 & $03)*10 + (i & $0F)
  'month
  i:=byte[@DS1307tm+5]
  tm_mon:=(i>>4 & $01)*10 + (i & $0F) - 1
  'year
  i:=byte[@DS1307tm+6]
  tm_year:=(i>>4 )*10 + (i & $0F) + 100  'adding 100 here because DS1307 time is based on year 2000, but this time is based on 1900
  'day of year
  tm_yday:=DayOfYear(tm_year,tm_mon,tm_mday)
  'dst
  tm_isdst:=-1  'unknown
  
  
PUB DayOfYear(year,mon,mday):yday
  'Gets the returns day of the year [0..365]
  if (NOT IsLeap(year))
    'normal year
     yday:=word[@NormDaysBeforeMon][mon]
  else
    'leap year
     yday:=word[@LeapDaysBeforeMon][mon]
  yday+=mday-1
  return yday   

PRI IN(register, n, pData)|i   'Input n bytes from DS1307 starting at address given by register and store starting at address given by pData

  START
  if (Write(DS1307)<>ACK)
    return false
  if (Write(register)<>ACK)
    return false
  START
  if (Write(DS1307+1)<>ACK)
    return false
  repeat i from 0 to n-1
    byte[pData+i]:=Read(ACK)
  STOP
  return true
    

PRI OUT(register, n, pData)|i    'Output n bytes to DS1307 to address starting at register from starting address given by pData 
   START
   if (Write(DS1307)<>ACK)
     return false
   if (Write(register)<>ACK)
     return false
   repeat i from 0 to n-1
     if (Write(byte[pData+i])<>ACK)
       return false
   STOP
   return (true)
            

PRI Start'(SCL) | SDA                   ' SDA goes HIGH to LOW with SCL HIGH
   outa[SCL]~~                         ' Initially drive SCL HIGH
   dira[SCL]~~
   outa[SDA]~~                         ' Initially drive SDA HIGH
   dira[SDA]~~
   outa[SDA]~                          ' Now drive SDA LOW
   outa[SCL]~                          ' Leave SCL LOW
  
PRI Stop'(SCL) | SDA                    ' SDA goes LOW to HIGH with SCL High
   outa[SCL]~~                         ' Drive SCL HIGH
   outa[SDA]~~                         '  then SDA HIGH
   dira[SCL]~                          ' Now let them float
   dira[SDA]~                          ' If pullups present, they'll stay HIGH

PRI Write(data) : ackbit '(SCL, data) : ackbit | SDA
'' Write i2c data.  Data byte is output MSB first, SDA data line is valid
'' only while the SCL line is HIGH.  Data is always 8 bits (+ ACK/NAK).
'' SDA is assumed LOW and SCL and SDA are both left in the LOW state.
   ackbit := 0 
   data <<= 24
   repeat 8                            ' Output data to SDA
      outa[SDA] := (data <-= 1) & 1
      outa[SCL]~~                      ' Toggle SCL from LOW to HIGH to LOW
      outa[SCL]~
   dira[SDA]~                          ' Set SDA to input for ACK/NAK
   outa[SCL]~~
   ackbit := ina[SDA]                  ' Sample SDA when SCL is HIGH
   outa[SCL]~
   outa[SDA]~                          ' Leave SDA driven LOW
   dira[SDA]~~

PRI Read(ackbit):data'(SCL, ackbit): data | SDA
'' Read in i2c data, Data byte is output MSB first, SDA data line is
'' valid only while the SCL line is HIGH.  SCL and SDA left in LOW state.
   data := 0
   dira[SDA]~                          ' Make SDA an input
   repeat 8                            ' Receive data from SDA
      outa[SCL]~~                      ' Sample SDA when SCL is HIGH
      data := (data << 1) | ina[SDA]
      outa[SCL]~
   outa[SDA] := ackbit                 ' Output ACK/NAK to SDA
   dira[SDA]~~
   outa[SCL]~~                         ' Toggle SCL from LOW to HIGH to LOW
   outa[SCL]~
   outa[SDA]~                          ' Leave SDA driven LOW


DAT

'this format is like the DS1302, but the order is slightly changed...
DS1307tm   byte                                            ' BIT7 | BIT6 | BIT5 | BIT4 | BIT3 | BIT2 | BIT1 | BIT0 |  RANGE   
DS1307seconds byte 0          'Address 0H                  ' CH   |      10 Seconds    |          Seconds          |   00-59
DS1307minutes byte 0          'Address 1H                  '      |      10 Minutes    |          Minutes          |   00-59 
DS1307hours   byte 0          'Address 2H                  '12/24 |  0   |10/PM | Hour |            HOUR           |   01-12 or 00-23
DS1307day     byte 0          'Address 3H                  '   0  |  0   |   0  |  0   |   0  |        DAY         |   1-7
DS1307date    byte 0          'Address 4H                  '   0  |  0   |  10 Date    |            DATE           |   01-31 
DS1307month   byte 0          'Address 5H                  '   0  |  0   |   0  | 10Mo |           MONTH           |   01-12
DS1307year    byte 0          'Address 6H                  '         10 Year           |           Year            |   00-99
DS1307control byte 0          'Address 7H                  ' Out  |  0   |   0  | SQWE |   0   |   0  | RS1 |  RS0 |
DS1307ram     byte 0[56]      'Address 08H..3FH


Time1 word         'Buffer for GMT
tm_sec   word 0                  '0 Seconds      [0-60] '60 allowed, but probably going to ignore leap seconds entirely
tm_min   word 0                  '1 Minutes      [0-59] 
tm_hour  word 0                  '2 Hours        [0-23] 
tm_mday  word 0                  '3 Day          [1-31] 
tm_mon   word 0                  '4 Month        [0-11] 
tm_year  word 0                  '5 Year - 1900         'same as in C language.  Should only need a byte for this, but an ill-configured DS1307 gives values >100 
tm_wday  word 0                  '6 Day of week  [0-6] 
tm_yday  word 0                  '7 Day of year  [0-365] 
tm_isdst word 0                  '8 DST          [-1/0/1] '-1=unknown, 0=not in DST, 1=in DST

Time2 word        'Buffer for User or Local Time
tm_sec2   word 0                  ' Seconds      [0-60] '60 allowed, but probably going to ignore leap seconds entirely
tm_min2   word 0                  ' Minutes      [0-59] 
tm_hour2  word 0                  ' Hours        [0-23] 
tm_mday2  word 0                  ' Day          [1-31] 
tm_mon2   word 0                  ' Month        [0-11] 
tm_year2  word 0                  ' Year - 1900         'same as in C language.  Should only need a byte for this, but an ill-configured DS1307 gives values >100 
tm_wday2  word 0                  ' Day of week  [0-6] 
tm_yday2  word 0                  ' Day of year  [0-365] 
tm_isdst2 word 0                  ' DST          [-1/0/1] '-1=unknown, 0=not in DST, 1=in DST

'Local Timezone info
tzname byte "EST",0
tzname0 byte "EST",0
tzname1 byte "EDT",0
timezone long SecondsPerHour*5      'EST
daylight long SecondsPerHour        'EDT offset

StrBuffer byte "Hello",0[40]  'string buffer
DaysOfWeek byte "Sun","Mon","Tue","Wed","Thu","Fri","Sat"
Months byte "Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"

'Table of # days that come before each month (0-12)
NormDaysBeforeMon word  0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334, 365   ' Normal years
LeapDaysBeforeMon word  0, 31, 60, 91, 121, 152, 182, 213, 244, 274, 305, 335, 366   ' Leap years

'Table for calculating DOW
Zeller byte 0, 3, 2, 5, 0, 3, 5, 1, 4, 6, 2, 4 


 